#include "sm3sum.h"

static void print_help()
{
	puts("Usage: sm3sum [OPTION] [FILE]");
	puts("Calculate or check SM3 (256-bit) checksum of a file.\n");
	puts("When [FILE] is -, read from standard input.");
	puts("  -c, --check       check the provided SM3 checksum and the input FILE's checksum\n");
	puts("      --help        display this help information");
	puts("      --version     output version information\n");
	puts("The SM3 sum is computed as described in GM/T 0004-2012.");
}


int main(int argc, char *argv[])
{
	sm3_file_ctx sm3CTX;
	memset(sm3CTX.filename, 0, PATH_LIMIT + 1);
	
	if (argc == 2)
	{
		if (strncmp(argv[1], "-v", 3) == 0 || strncmp(argv[1], "--version", 10) == 0)
		{
			printf("sm3sum version %s\n", VERSION);
			return 0;
		}

		if (strncmp(argv[1], "-h", 3) == 0 || strncmp(argv[1], "--help", 7) == 0)
		{
			print_help();
			return 0;
		}

		if (strncmp(argv[1], "-", 2) == 0)
		{
			sm3CTX.filename[0] = '-';
		}
		else if ( strlen(argv[1]) > PATH_LIMIT )
		{
			puts("Length of input filename exceeds!");
			return -1;
		}
		else // sm3sum [FILE]
		{
		#if defined(__GNUC__)
			#pragma GCC diagnostic push
			#pragma GCC diagnostic ignored "-Wstringop-truncation"
		#endif
			// Size already check. Ignore -Wstringop-truncation
			strncpy( sm3CTX.filename, argv[1], strlen(argv[1]) );
		#if defined(__GNUC__)
			#pragma GCC diagnostic pop
		#endif
		}

		calc_file_hash(&sm3CTX);
		print_file_sm3(&sm3CTX);
		return 0;
	}

	if (argc == 4)
	{
		if ( strncmp(argv[1], "-c", 3) == 0 || strncmp(argv[1], "--check", 8) == 0 )  // sm3sum -c [FILE] SM3SUM
		{
			if ( strlen(argv[2]) > PATH_LIMIT )
			{
				puts("Length of input filename exceeds!");
				return -1;
			}
			if ( strlen(argv[3]) != SM3_DIGEST_LENGTH * 2 )
			{
				puts("Length of input sm3 hash exceeds!");
				return -1;
			}
		#if defined(__GNUC__)
			#pragma GCC diagnostic push
			#pragma GCC diagnostic ignored "-Wstringop-truncation"
		#endif
			// Size already check. Ignore -Wstringop-truncation
			strncpy( sm3CTX.filename, argv[2], strlen(argv[2]) );
		#if defined(__GNUC__)
			#pragma GCC diagnostic pop
		#endif
			sm3str_to_int(argv[3], sm3CTX.expected_sm3);
			
			return check_file_hash(&sm3CTX);
		}	
	}
	
	print_help();
	return 0;
}
