#ifndef _SM3_H_
#define _SM3_H_

#include<stdio.h>
#include<stdlib.h>
#include<stdint.h>
#include<string.h>
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <string.h>
#include <stdlib.h>

typedef uint64_t	u8;
typedef uint32_t	u4;
typedef uint16_t	u2;
typedef uint8_t		u1;

// All in bytes
#define SM3_BLOCK_SIZE      64
#define SM3_DIGEST_LENGTH	32
#define SM3_HMAC_SIZE	    32
#define SM3_CBLOCK	        (SM3_BLOCK_SIZE)

typedef struct _SM3_CTX
{
	uint32_t   digest[8];
	uint32_t   nblocks;                // How many blocks have been compressed. 
	uint8_t    block[SM3_BLOCK_SIZE];  // Store the rest data which not enougn one block
	uint32_t   num;                    // Record how many bytes in block[]

} SM3_CTX;


typedef struct _SM3_HMAC_CTX
{
	SM3_CTX sm3_ctx;
	uint8_t key[SM3_BLOCK_SIZE];
} SM3_HMAC_CTX;


void	sm3_init		(SM3_CTX *ctx);
void	sm3_update		(SM3_CTX *ctx, const uint8_t *data, size_t data_len);
void	sm3_final		(SM3_CTX *ctx, uint8_t digest[SM3_DIGEST_LENGTH]);
void	sm3_compress	(u4 digest[8], const uint8_t* block, u4 nblocks);
size_t	sm3				(const uint8_t *data, size_t datalen, uint8_t dgst[SM3_DIGEST_LENGTH]);


void	sm3_hmac_init	(SM3_HMAC_CTX * ctx, uint8_t * key, size_t key_len);
void	sm3_hmac_update	(SM3_HMAC_CTX * ctx, const uint8_t * data, size_t data_len);
void	sm3_hmac_final	(SM3_HMAC_CTX * ctx, uint8_t mac[SM3_HMAC_SIZE]);
size_t	sm3_hmac		(const uint8_t * data, size_t datalen, uint8_t * key, size_t key_len, uint8_t mac[SM3_HMAC_SIZE]);

#endif // end of sm3.h

